/*
  Kinesis ergonomic keyboard firmware replacement

  Copyright 2012 Chris Andreae (chris (at) andreae.gen.nz)

  Licensed under the GNU GPL v2 (see GPL2.txt).

  ==========================

  If built for V-USB, this program includes library and sample code from:
	 V-USB, (C) Objective Development Software GmbH
	 Licensed under the GNU GPL v2 (see GPL2.txt)

  ==========================

  If built for LUFA, this program includes library and sample code from:
			 LUFA Library
	 Copyright (C) Dean Camera, 2011.

  dean [at] fourwalledcubicle [dot] com
		   www.lufa-lib.org

  Copyright 2011  Dean Camera (dean [at] fourwalledcubicle [dot] com)

  Permission to use, copy, modify, distribute, and sell this
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in
  all copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaim all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

#ifndef __KINESIS_H
#define __KINESIS_H

#include "keystate.h"

// Use Obdev's free usb id for keyboards discriminated by serial number
#define USB_VENDOR_ID  0x16c0
#define USB_PRODUCT_ID 0x27db

// But define different descriptive strings
#define USB_MANUFACTURER_STRING L"andreae.gen.nz"
#define USB_PRODUCT_STRING L"Programmable USB Keyboard"
#define USB_SERIAL_NUMBER_STRING L"andreae.gen.nz:kinesis"

// Unique identifier representing this keyboard's layout and
// definition of logical_keycode values.  Is reported to the
// configuration program over USB to identify the layout.
#define LAYOUT_ID 1

/* Storage layout */
#define CONSTANT_STORAGE           avr_pgm
#define MAPPING_STORAGE            avr_eeprom
#define SAVED_MAPPING_STORAGE      avr_eeprom
#define SAVED_MAPPING_COUNT        256          // 2-byte entries
#define MACRO_INDEX_STORAGE        avr_eeprom
#define MACRO_INDEX_COUNT          50           // 6-byte entries
#define MACROS_STORAGE             i2c_eeprom
#define MACROS_SIZE                1024
#define PROGRAM_STORAGE            i2c_eeprom
#define PROGRAM_SIZE               1024
#define PROGRAM_COUNT              6

/* Kinesis Matrix */

// The Kinesis has few physical keys, so has a "keypad layer" toggle
// which has a separate binding for each key. This provides access not
// only to the numpad, but to the windows and menu keys.

#define NUM_PHYSICAL_KEYS 86
#define NUM_LOGICAL_KEYS  NUM_PHYSICAL_KEYS * 2
#define KEYPAD_LAYER_SIZE NUM_PHYSICAL_KEYS

#define MATRIX_COLS 10 // 8 demultiplexer selected matrix columns, and two direct button inputs (pins 5 and 6)
#define MATRIX_ROWS 16 // 2 74LS138 1-of-8 demultiplexers

// Logical keys we have
enum logical_keys {
	LOGICAL_KEY_PROGRAM, // PG
	LOGICAL_KEY_KEYPAD, // KY
	LOGICAL_KEY_A,
	LOGICAL_KEY_B,
	LOGICAL_KEY_C,
	LOGICAL_KEY_D,
	LOGICAL_KEY_E,
	LOGICAL_KEY_F,
	LOGICAL_KEY_G,
	LOGICAL_KEY_H,
	LOGICAL_KEY_I,
	LOGICAL_KEY_J,
	LOGICAL_KEY_K,
	LOGICAL_KEY_L,
	LOGICAL_KEY_M,
	LOGICAL_KEY_N,
	LOGICAL_KEY_O,
	LOGICAL_KEY_P,
	LOGICAL_KEY_Q,
	LOGICAL_KEY_R,
	LOGICAL_KEY_S,
	LOGICAL_KEY_T,
	LOGICAL_KEY_U,
	LOGICAL_KEY_V,
	LOGICAL_KEY_W,
	LOGICAL_KEY_X,
	LOGICAL_KEY_Y,
	LOGICAL_KEY_Z,
	LOGICAL_KEY_1,
	LOGICAL_KEY_2,
	LOGICAL_KEY_3,
	LOGICAL_KEY_4,
	LOGICAL_KEY_5,
	LOGICAL_KEY_6,
	LOGICAL_KEY_7,
	LOGICAL_KEY_8,
	LOGICAL_KEY_9,
	LOGICAL_KEY_0,
	LOGICAL_KEY_F1,
	LOGICAL_KEY_F2,
	LOGICAL_KEY_F3,
	LOGICAL_KEY_F4,
	LOGICAL_KEY_F5,
	LOGICAL_KEY_F6,
	LOGICAL_KEY_F7,
	LOGICAL_KEY_F8,
	LOGICAL_KEY_F9,
	LOGICAL_KEY_F10,
	LOGICAL_KEY_F11,
	LOGICAL_KEY_F12,
	LOGICAL_KEY_EQUALS,
	LOGICAL_KEY_TAB,
	LOGICAL_KEY_ESC,
	LOGICAL_KEY_CAPSLOCK,
	LOGICAL_KEY_LSHIFT,
	LOGICAL_KEY_UP,
	LOGICAL_KEY_DOWN,
	LOGICAL_KEY_LEFT,
	LOGICAL_KEY_RIGHT,
	LOGICAL_KEY_LSQUARE,
	LOGICAL_KEY_RSQUARE,
	LOGICAL_KEY_SCROLLLOCK,
	LOGICAL_KEY_PRINTSCREEN,
	LOGICAL_KEY_INTERNATIONAL, // insert in keypad layer
	LOGICAL_KEY_BACKTICK,
	LOGICAL_KEY_HYPHEN,
	LOGICAL_KEY_PAUSE,
	LOGICAL_KEY_END,
	LOGICAL_KEY_HOME,
	LOGICAL_KEY_DELETE,
	LOGICAL_KEY_BACKSPACE,
	LOGICAL_KEY_L_ALT,
	LOGICAL_KEY_L_CTRL,
	LOGICAL_KEY_BACKSLASH,
	LOGICAL_KEY_PGDN,
	LOGICAL_KEY_PGUP,
	LOGICAL_KEY_ENTER,
	LOGICAL_KEY_SPACE,
	LOGICAL_KEY_R_ALT,
	LOGICAL_KEY_R_CTRL,
	LOGICAL_KEY_SEMICOLON,
	LOGICAL_KEY_QUOTE,
	LOGICAL_KEY_COMMA,
	LOGICAL_KEY_PERIOD,
	LOGICAL_KEY_SLASH,
	LOGICAL_KEY_RSHIFT,
	// The keypad layer duplicates the previous 86 keys
};

// Which keys to use for special in-built combinations
#define SPECIAL_HKEY_MACRO_RECORD  HID_KEYBOARD_SC_F11
#define SPECIAL_HKEY_REMAP         HID_KEYBOARD_SC_F12
#define SPECIAL_HKEY_REBOOT        HID_KEYBOARD_SC_PRINT_SCREEN
#define SPECIAL_HKEY_TOGGLE_BUZZER HID_KEYBOARD_SC_BACKSLASH_AND_PIPE
#define SPECIAL_HKEY_RESET_CONFIG  HID_KEYBOARD_SC_F7
#define SPECIAL_HKEY_RESET_FULLY   HID_KEYBOARD_SC_LEFT_SHIFT

extern const logical_keycode matrix_to_logical_map[MATRIX_ROWS][MATRIX_COLS] PROGMEM;

/* For each key, maps an index position to a default HID key code. */
/* stored in flash. */
extern const hid_keycode logical_to_hid_map_default[NUM_LOGICAL_KEYS] PROGMEM;


//////////////////////////////////////////////////////////////////////////////////////
// Kinesis board layout

// 74LS138 demultiplexer matrix select output:
//  A-C connect to the address pins 1-3 of the LS138s, externally pulled up.
//  P138SEL connects to pin 4 (negated input) of the first and pin 6 of the second,
//  to select between them, also externally pulled up.
//  The opposing pins are tied to VCC and ground respectively.
//  Pin 10 ("WP") is connected to pin 5 of both demultiplexers, and must be pulled low to enable them.
//    A (pin 39)
//    B (pin 38)
//    C (pin 37)
//    P138SEL (pin 36)

// Key input:
//  Pressing the key pulls low. I don't think there's an external pull-up, so am using internal pull-ups.
//    keypad (pin 5)
//    program (pin 6)
//    matrix output (pins 21-28)
//    foot switch 1 (pin 11)
//    foot switch 2 (pin 17) - (note that these may differ on the 3-foot-switch Advantage models)

// output:
//    4 LEDs (pins 1-4, sink current to enable) (num=1, scrolllock=2, kpd=3, caps=4)
//    buzzer (pin 32)

// EEPROM access:
//  have access to up to two AT24C164 2kb serial eeproms via pins 7 and
//  8, connected to SCL and SDA respectively with external pull-ups.
//  The eeproms have their addresse lines tied to 000 and 001 respectively.
//         WP  (write protect) (pin10) = connected to WP pin of eeprom, must be low to enable.
//                                       (Note that this is also connected to the demultiplexers: most convenient to just tie it permanently to ground)
//         SCL (serial eeprom clock line) (pin7)
//         SDA (serial eeprom data line) (pin8)

// static configuration:
//   pin 20 - GND
//   pin 40 - VCC

// unused pins:
//   pin 12 - PS/2 clock
//   pin 13 - PS/2 data
////////////////////////////////////////////////////////////////////////////////////////

#if defined(__AVR_ATmega16A__)
// Original board layout (v1)

// PD4 = USB D-
// PD2 = USB D+

#define MATRIX_PORT PORTB
#define MATRIX_DDR  DDRB
#define MATRIX_SELECT_A (1<<0)
#define MATRIX_SELECT_B (1<<1)
#define MATRIX_SELECT_C (1<<2)
#define MATRIX_SELECT_P138SEL (1<<3)
#define MATRIX_MASK (MATRIX_SELECT_A | MATRIX_SELECT_B | MATRIX_SELECT_C | MATRIX_SELECT_P138SEL)

// keypad key
#define INPUT_PIN5_PIN  PIND
#define INPUT_PIN5_PORT PORTD
#define INPUT_PIN5_DDR  DDRD
#define INPUT_PIN5 (1<<3)

// program key
#define INPUT_PIN6_PIN  PINC
#define INPUT_PIN6_PORT PORTC
#define INPUT_PIN6_DDR  DDRC
#define INPUT_PIN6 (1<<0)

// FS1 and FS2 are PC5 and PC6

#define INPUT_REST_PIN  PINA
#define INPUT_REST_PORT PORTA
#define INPUT_REST_DDR  DDRA
#define INPUT_REST (0xFF)

#define LED_PORT PORTD
#define LED_DDR  DDRD
#define LED_CAPS (1<<4)
#define LED_NUMLOCK (1<<7)
#define LED_SCROLLLOCK (1<<6)
#define LED_KEYPAD (1<<5)
#define ALL_LEDS (LED_CAPS | LED_NUMLOCK | LED_SCROLLLOCK | LED_KEYPAD)

#define USE_BUZZER 0 // new buzzer code uses OC2 pin, not compatible with this board layout

#define BUZZER_PORT PORTC
#define BUZZER_DDR DDRC
#define BUZZER (1<<3)

#define BITBANG_TWI // original board bitbangs TWI, doesn't use hardware I2C pins

#define EEPROM_PORT PORTC
#define EEPROM_DDR  DDRC
#define EEPROM_PIN  PINC
#define EEPROM_SCL (1<<1)
#define EEPROM_SDA (1<<2)

#elif defined(__AVR_ATmega32__)
// New board layout (v2)

// PD4 = USB D-
// PD2 = USB D+

#define MATRIX_PORT PORTB
#define MATRIX_DDR  DDRB
#define MATRIX_SELECT_A (1<<0)
#define MATRIX_SELECT_B (1<<1)
#define MATRIX_SELECT_C (1<<2)
#define MATRIX_SELECT_P138SEL (1<<3)
#define MATRIX_MASK (MATRIX_SELECT_A | MATRIX_SELECT_B | MATRIX_SELECT_C | MATRIX_SELECT_P138SEL)

// keypad key
#define INPUT_PIN5_PIN  PINC
#define INPUT_PIN5_PORT PORTC
#define INPUT_PIN5_DDR  DDRC
#define INPUT_PIN5 (1<<3)

// program key
#define INPUT_PIN6_PIN  PINC
#define INPUT_PIN6_PORT PORTC
#define INPUT_PIN6_DDR  DDRC
#define INPUT_PIN6 (1<<2)

// FS1 and FS2 are PC5 and PC6

#define INPUT_REST_PIN  PINA
#define INPUT_REST_PORT PORTA
#define INPUT_REST_DDR  DDRA
#define INPUT_REST (0xFF)

#define LED_PORT PORTD
#define LED_DDR  DDRD
#define LED_CAPS (1<<3)
#define LED_NUMLOCK (1<<6)
#define LED_SCROLLLOCK (1<<5)
#define LED_KEYPAD (1<<4)
#define ALL_LEDS (LED_CAPS | LED_NUMLOCK | LED_SCROLLLOCK | LED_KEYPAD)

#define USE_BUZZER 1

#define BUZZER_PORT PORTD
#define BUZZER_DDR DDRD
#define BUZZER (1<<7)

#define EEPROM_PORT PORTC
#define EEPROM_DDR  DDRC
#define EEPROM_PIN  PINC
#define EEPROM_SCL (1<<0)
#define EEPROM_SDA (1<<1)

#else
#error Ports not yet defined for this microcontroller
#endif

void ports_init(void);

/**
 * Gets the current physical input for a given physical position
 */
void matrix_select_row(uint8_t matrix_row);
uint8_t matrix_read_column(uint8_t matrix_column);

/* Macros: */
/** LED mask for the library LED driver, to indicate that the USB interface is not ready. */
#define LEDMASK_USB_NOTREADY     (LED_KEYPAD | LED_NUMLOCK)

/** LED mask for the library LED driver, to indicate that the USB interface is enumerating. */
#define LEDMASK_USB_ENUMERATING  (LED_KEYPAD | LED_CAPS)

/** LED mask for the library LED driver, to indicate that the USB interface is ready. */
#define LEDMASK_USB_READY        0

/** LED mask for the library LED driver, to indicate that an error has occurred in the USB interface. */
#define LEDMASK_USB_ERROR        (LED_KEYPAD | LED_CAPS | LED_NUMLOCK)

#define LEDMASK_CAPS      LED_CAPS
#define LEDMASK_NUMLOCK   LED_NUMLOCK
#define LEDMASK_SCROLLLOCK LED_SCROLLLOCK
#define LEDMASK_KEYPAD    LED_KEYPAD

#define LEDMASK_PROGRAMMING_SRC (LED_CAPS|LED_SCROLLLOCK)
#define LEDMASK_PROGRAMMING_DST (LED_NUMLOCK|LED_SCROLLLOCK)
#define LEDMASK_MACRO_TRIGGER   (LED_CAPS|LED_NUMLOCK)
#define LEDMASK_MACRO_RECORD    (LED_CAPS|LED_NUMLOCK|LED_SCROLLLOCK)
#define LEDMASK_ALL ALL_LEDS
#define LEDMASK_NONE 0

void set_all_leds(uint8_t led_mask);

void test_leds(void);

#endif // __KINESIS_H
